/*
* CORE.CPP                                                Version 1.0
*
* !!! THIS MODULE IS NOT OPTIONAL !!!
*
* CORE MODULE. Includes smx configuration table, main entry point,
* initialization, idle task, timeout task, exit task, C versions of
* interrupt service routines.
*
* Copyright (c) 1990-1993 Micro Digital Associates, Inc. 
*               All Rights Reserved.
*
* Author: Ralph Moore
* Update: Rick Evans
* Updated module for SMX++.
*
***********************************************************************/

#include    "core.h"
#include    "conf.h"
#include    <bios.h>
#include    <conio.h>
#include    <stdlib.h>
#include    <stdio.h>
#include    <ddemo.hpp>

#ifdef   __TURBOC__
 #define _bios_keybrd bioskey
 #undef  outp   /*(forces fct rather than macro to avoid warning)*/
#endif

#define  DIVR(x,y) ((2*((x)%(y)) < (y)) ? (x)/(y) : ((x)/(y))+1)

/* function prototypes */

#if __cplusplus
 extern "C" {
#endif

void        ainit(void);
void        idle_main(void);
void        idle_state(void);
void near   keep_time(void);
void        sp_exit(void);
void        sp_init(void);
void far    tick_int(void);
void interrupt far tick_sim(void);
void        timeout_init(void);
void        timeout_main(void);

#if __cplusplus
}
#endif

/* variables */

Task*       ExitxTp;             /* end of operation task class ptr */
Task*       TimeOutTp;           /* end of operation task class ptr */
Task*       IdleTp;              /* end of operation task class ptr */
Sem*        InClibSp;            /* in the C lib */
Sem*        InDosSp;             /* in MS-DOS call */
Sem*        TsSp;
Counter*    TicksCp;             /* ticks event queue */

TCB_PTR     near  exitx;
SCB_PTR     near  in_dos;        /* in MS-DOS call used for ASM code */ 
word        near  ictr;          /* idle tick counter */
struct DAR  near  ndar;          /* ndar structure */
CB_PTR      near  null_cb;       /* null control block */
word        near  octr;          /* overhead tick counter */
int         near  pcticks;       /* counts ticks for tick simulation */
word              speed_up;      /* tick speed up factor */
word        near  tick_count;    /* counts ticks for tick speed up */
word        near  tick_ctr;      /* counts ticks for ts signalling */
word        near  wctr;          /* (useful) work tick counter */

/* array pointers to aid debugging. (NOTE: array sizes must be specified
   for code debuggers.) */

struct EREC     (far *e)[EB_SIZE];        /* eb array ptr */
struct LQ_CELL (near *l)[LQ_SIZE];        /* lq array ptr */
struct XCB     (near *r)[MAX_PRIORITY+1]; /* rq array ptr */

/* old isr pointers */

#ifndef __ZTC__
void (interrupt far * near dos_call)(void);  /* --> DOS call code */
void (interrupt far * near old_atdisk)(void);/* --> XT disk isr */
void (interrupt far * near old_timer)(void); /* --> PC timer isr */
void (interrupt far * near old_xtdisk)(void);/* --> AT disk isr */
#else
void far * near dos_call;                    /* --> DOS call code */
void far * near old_atdisk;                  /* --> XT disk isr */
void far * near old_timer;                   /* --> PC timer isr */
void far * near old_xtdisk;                  /* --> AT disk isr */
#endif

/* default configuration values -- see ddconf.h */

struct CONF near   cf =  {
         /* max_priority */   MAX_PRIORITY,
         /* lq_size      */   LQ_SIZE,
         /* sec          */   SEC,
         /* num_blocks   */   NUM_BLOCKS,
         /* num_buckets  */   NUM_BUCKETS,
         /* num_msgs     */   NUM_MSGS,
         /* num_pools    */   NUM_POOLS,
         /* num_qlevels  */   NUM_QLEVELS,
         /* num_pipes    */   NUM_PIPES,
         /* num_tasks    */   NUM_TASKS,
         /* stack_size   */   STACK_SIZE,
         /* num_stacks   */   NUM_STACKS,
         /* eb_size      */   EB_SIZE,
         /* sp_min       */   &bstack,
         /* sp_max       */   &estack,
         /* nheap_min    */   &bnheap,
         /* nheap_max    */   &enheap,
         /* ndar_min     */   &bndar, 
         /* ndar_max     */   &endar, 
         /* fheap_min    */   0,
         /* fheap_max    */   0};


/******* MAIN
* This is the entry point from the C compiler startup code.
***********************************************************************/

void _cdecl main(int argc, char *argv[])
{
   if(argc > 1)
      speed_up = atoi(argv[1]);
   else
      speed_up = 1;
   cf.sec *= speed_up;
   coldstart();
}


/******* APPLICATION INITIALIZATION
* This is the starting point for the task environment. go_smx() (called
* at the end of coldstart) assumes ainit() to exist. This function is
* executed under the idle task. Interrupts should be masked until 
* initialization is complete, since smx calls (e.g. go_smx()) set the
* interrupt enable flag, IF. At the end of ainit(), the idle task code
* is changed to idle_main() and its priority is set to MIN. 
***********************************************************************/

void _cdecl ainit(void)
{
   word  i, val;

   // setup c++ interface for smx++.
   // This will allow you to access these objects through their
   // class methods.

   IdleTp      = new Task( idle );
   TimeOutTp   = new Task( timeout );
   TsSp        = new Sem( ts );
   TicksCp     = new Counter( ticks );

   /* debug pointers, use: *e to examine error buffer
                           *l to examine lsr queue
                           *r to examine ready queue */
   #ifdef   TM
    e = (struct EREC (far *)[EB_SIZE])ebi;
    l = (struct LQ_CELL (near *)[LQ_SIZE])lqi;
    r = (struct XCB (near *)[MAX_PRIORITY+1])rq;
   #endif

   sp_init();        /* initialize for smxProbe */
   kbd_init();

   if(xerrno) exitx_main(xerrno); /* (so error message will be seen) */

   clr_screen(LIGHTGRAY, BLACK);
   bios_cursor_off();
   unlock(ct);         /* so can use smxProbe in ainit */

   /* DON'T USE SMXPROBE BEFORE HERE. */

   ndar.dnext = cf.ndar_min;   /* set up ndar for create_pool() calls */
   ndar.dmax  = cf.ndar_max;

   #ifdef DEBUG
      old_timer = exch_vect(8, tick_sim);
   #else
      old_timer = exch_vect(8, tick);
   #endif

   if(speed_up > 1)
      {
      _disable();
      outp(0x43, 0x36);     /* set timer 0 to mode 3 */
      for (i=0; i<2; i++);	 /* delay */
      val = (word)(65536L/speed_up);
      outp(0x40, (val & 0x00FF)); /* load LSB of val */
      for (i=0; i<2; i++);			 /* delay */
      outp(0x40, (val >> 8));	    /* load MSB of val */
      _enable();
      }

   // The exitx task is used at the C level. By creating it as
   // below it can be accessed through C and C++.

   exitx = create_task((void (*)())exitx_main, MAX, 0);
   ExitxTp = new Task( exitx );

   // Allows C and C++ access to the InDos semaphor.

   in_dos = create_sem(1,0); /* used for ASM code */
   InDosSp = new Sem( in_dos );
   InDosSp->Signal();

   /* NOTE: The following are inhibited for DEBUG because not necessary
      and cause difficulty if exit from debugger without restoring. */

   #ifndef  DEBUG
      dos_call = exch_vect(0x21, dos_trap);
      old_atdisk = exch_vect(0x76, new_atdisk);
      old_xtdisk = exch_vect(0x0d, new_xtdisk);
   #endif

   // This semaphor is only accessed through C++.

   InClibSp = new Sem(1,0); /* to be used for non-reen. C lib. calls */
   InClibSp->Signal();

   /* init application specific info */
   appl_init();

   /* Build handle table for smxProbe. (These are nop's if smxProbe is
      not present.) */

   BUILD_HT(in_dos, "in_dos");
// BUILD_HT(in_clib,"in_clib");
   BUILD_HT(exitx,  "exitx");
   BUILD_HT((VOID_PTR)keep_time, "keep_time");

   IdleTp->Start( idle_main, MIN );  /* become normal idle task */
}

      
/******* IDLE
* The idle task runs when there is no other task to run. It can be used
* to compute processor utilization as well as to gather other informa-
* tion of interest. It can also be used to perform confidence tests
* such as heap checks, stack checks, etc.
***********************************************************************/

void _cdecl idle_main()
{
   static char buffer[10];
   static int  percent_idle;  /* % of time in idle task */ 
   static int  percent_ovh;   /* % of time in smx */
   static int  percent_work;  /* % of time doing useful work */
   int         total;


   busy = 1;         /* system is idle whenever this task is running */
   IdleTp->Hook( idle_state ); /* set busy flag to idle whenever resume idle */
   IdleTp->UnLock(); /* permit this task to be preempted */

   while(TRUE)
      {
      total = octr + ictr + wctr;   /* total ticks since last profile */

      /* Calculate and output percentages. NOTE: The following could be
         done more easily using floating point numbers. However, in many
         applications, it is undesirable to introduce floating point. */

      if(total >= 30)               /* time for new profile? */
         {
         if(percent_work == 0)      /* first time */
            {
            percent_ovh  = DIVR(1000*octr,total);
            percent_idle = DIVR(1000*ictr,total);
            percent_work = DIVR(1000*wctr,total);
            }
         else                       /* subsequent times -- average */
            {
            percent_ovh  = DIVR((DIVR(1000*octr,total) + 3*percent_ovh),4);
            percent_idle = DIVR((DIVR(1000*ictr,total) + 3*percent_idle),4);
            percent_work = DIVR((DIVR(1000*wctr,total) + 3*percent_work),4);
            }

         ictr = octr = wctr = 0;    /* clear counters */

         InClibSp->Test(INF);
         sprintf(buffer, "%3d", percent_idle/10);
         wr_string(9,15,YELLOW,BLACK,!BLINK,buffer);
         sprintf(buffer, ".%1d", percent_idle%10);
         wr_string(12,15,YELLOW,BLACK,!BLINK,buffer);

         sprintf(buffer, "%3d", percent_work/10);
         wr_string(9,16,YELLOW,BLACK,!BLINK,buffer);
         sprintf(buffer, ".%1d", percent_work%10);
         wr_string(12,16,YELLOW,BLACK,!BLINK,buffer);

         sprintf(buffer, "%3d", percent_ovh/10);
         wr_string(9,17,YELLOW,BLACK,!BLINK,buffer);
         sprintf(buffer, ".%1d", percent_ovh%10);
         wr_string(12,17,YELLOW,BLACK,!BLINK,buffer);
         InClibSp->Signal();
         }

      /* simulate tick for operation with code debugger */

      #ifdef DEBUG
       if(pcticks)
          {
          pcticks--;  /* decrement before task switch if timeslicing */
          #ifdef __ZTC__
           tick_int();
          #else
           (*tick)();
          #endif
          }
      #endif
      }
}


/******* TIMEOUT
* This task is activated by a signal to the ts semaphore. (See the
* keep_time lsr.) It searches for tasks which have timed out (i.e. 
* task timer > 0 and <= etime.) When found, such a task is put into
* the ready queue and its timer is cleared. This is repeated until all
* timers have been checked. go_smx() expects timeout_init() to exist.
***********************************************************************/

void _cdecl timeout_init()
{
   self->fun = conv_code_ptr(timeout_main);
   self->priority = HI;

   TsSp->TestStop();
}

void _cdecl timeout_main()
{
   register word     tn;
   register TCB_PTR  task;

   TimeOutTp->UnLock();
   
   show_etime();     /* optional */

   for( tn = 0, task = tcbi ; task <= tcbx ; tn++, task++)
      if(timer[tn] != 0L)
         resume_to(task);

/* If the most significant bit of etime is set, clear it and clear the
   most significant bit in all timers. NOTE: ok to use etime directly
   since only one bit is being tested and changed. */

   if(etime & 0x80000000L)
      {
      for( tn = 0, task = tcbi ; task <= tcbx ; tn++, task++)
         timer[tn] &= 0x7FFFFFFFL;
      etime &= 0x7FFFFFFFL;
      }

   TsSp->TestStop();
}


/****** EXIT SMX
* This function may be executed either by directly calling it or by
* starting the exitx task. Its purpose is to ensure a clean exit. In
* order to avoid exiting during a DOS call, the exitx task will wait
* up to 2 seconds for DOS to finish.
***********************************************************************/   

void _cdecl exitx_main(word errno)
{
   word  i;

   InDosSp->Test(2*SEC);  /* wait up to 2 seconds for DOS op to finish */

   if(speed_up > 1)  /* restore normal tick rate */
      {
      _disable();
      outp(0x43, 0x36);    /* set timer 0 to mode 3 */
      for (i=0; i<2; i++);	/* delay */
      outp(0x40, 0);       /* load LSB of 0 */
      for (i=0; i<2; i++); /* delay */
      outp(0x40, 0);	      /* load MSB of 0 */
      _enable();
      }

   if(old_timer) exch_vect(8, old_timer);

   #ifndef  DEBUG
      if(dos_call) exch_vect(0x21, dos_call);
      if(old_atdisk) exch_vect(0x76, old_atdisk);
      if(old_xtdisk) exch_vect(0x0d, old_xtdisk);
   #endif

   /* shut down process */
   delete DPP;
	delete InDosSp;
	delete InClibSp;

   quit_appl();
   kbd_exit();
   sp_exit();

   if(errno == NULL_PTR_REF)
      null_cb->fl = 0;   /* clear null error before msc calls */

   #ifdef DEBUG
    wr_string(0,22,LIGHTGRAY,BLACK,!BLINK,xerrmsg[errno]);
   #endif

   cursor_off();
   bios_cursor_on();

   while(_bios_keybrd(1))  /* empty bios keyboard buffer */
      _bios_keybrd(0);

   if(errno)
      _exit(errno);
   else
      if(_doserrno)
         _exit(_doserrno+100);
      else
         _exit(0);
}


/******* INTERRUPT SERVICE ROUTINES
* These are C versions of basic interrupt service routines. See bios.asm
* for equivalent assembly versions. C_ISRS in conf.h controls which
* is used.
***********************************************************************/

#ifndef __ZTC__		/* Zortech must use assembly ISR's	*/

/* The tick_sim isr does not require ENTER_ISR & EXIT_ISR because it 
   does no invokes and it has higher priority than any interrupt using
   enter_isr and exit_isr. */

void interrupt far tick_sim()
{
   if( speed_up > 1)
   {
      if (++tick_count >= speed_up)
      {
         tick_count = 0;
         (* old_timer)();
      }
      else
         outp(0x20, 0x20);  /* EOI to interrupt controller */
   }
   else
      (* old_timer)();

   pcticks++;
}

#if C_ISRS == 1   /* Enable C isr's and lsr's */ 

/* examples of "encapsulated" interrupts */

void interrupt far new_atdisk()
{
   ENTER_ISR();
   (* old_atdisk)();
   EXIT_ISR();
}

/* NOTE: Encapsulating the following interrupt for an AT will permit
   operation with LANtastic (if it is using irq5). */

void interrupt far new_xtdisk()
{
   ENTER_ISR();
   (* old_xtdisk)();
   EXIT_ISR();
}

void interrupt far tick()
{
   ENTER_ISR();
   INVOKE(keep_time, 0);

   /* Increment profile counters. NOTE: Profiling will be more accurate
      if the following is done for an asynchronous interrupt. */

   if(busy > 1)      
      wctr++;
   else if(busy == 1)
      ictr++;
   else
      octr++;

   /* Chain to the BIOS timer isr. */

  #ifndef DEBUG
	if(speed_up > 1)
      if (++tick_count >= speed_up)
      {
         tick_count = 0;
         (* old_timer)();
      }
		else
         outp(0x20, 0x20);  /* EOI to interrupt controller */
   else
      (* old_timer)();
#endif

   EXIT_ISR();
}

/******* KEEP TIME LSR
* NOTE: If using the following lsr and using Microsoft C, add /NT_TEXT
* to the compiler command line for this module. For Turbo C, add -zC_TEXT.
***********************************************************************/
       
void _cdecl near keep_time()
{
   #if TIME_SLICE == 1
    if(rq->fl)
       bump_task((TCB_PTR)rq->fl,0);
   #endif

   etime++;
   TicksCp->Signal();
   if(++tick_ctr >= cf.sec)
      {
      tick_ctr = 0;
      stime++;
      TsSp->Signal();
      }
}

#endif /* C_ISRS */
#endif /* __ZTC__ */


/******* HEAP TRANSITION ROUTINES
* The following functions convert compiler heap calls to smx heap calls
* at link time. (See xheap.h for macros to convert at compile time.)
***********************************************************************/

#ifdef   __TURBOC__

void *calloc(size_t nitems, size_t size)
{
   #if !LARGE_DATA
	 return (_ncallocx(nitems, size));
   #else
	 return (_fcallocx(nitems, size));
   #endif
}

void free(void *block)
{
   #if !LARGE_DATA
    _nfreex(block);
   #else
    _ffreex(block);
   #endif
}

#ifdef   MSC6

void _ffree(void far *block)
{
   _ffreex(block);
}

void _nfree(void near *block)
{
   _nfreex(block);
}
#endif

/* malloc's are performed by compiler startup code for standard i/o
   environment space, and other puposes. At the time, smx heaps have
   not been initialized. To handle this, two approaches are shown:
   (1) simple allocation from a fixed-size buffer above ndar (so it
       won't be cleared.)
   (2) allocation from DOS. (commented out)
   The second is more flexible, but will not work for a small data
   model. The best solution is to modify the compiler startup code. */

void *malloc(size_t size)
{
   static char *this_pos, *next_pos;
//   static union REGS regs;

   next_pos  = (char *)&temp_heap;

   if(!smx_init)  /* if smx heaps not initialized */
      {
      this_pos = next_pos;
      next_pos += size;
      return(this_pos);
//		size >>= 4;    /* convert size to paragraphs */
//		size++;		
//		regs.x.bx = size; /* make DOS memory allocation call */
//		regs.h.ah = 0x48; 	
//		intdos(&regs, &regs); 
//      return(regs.x.cflag ? NULL : MK_FP(regs.x.ax, 0));
   	}

   #if !LARGE_DATA
    return(_nmallocx(size));
   #else
    return(*_fmallocx(size));
   #endif
}
#endif   

/******* SMXPROBE INVOKE FUNCTION
* The following function permits invoking smxProbe from within a code
* debugger.
***********************************************************************/

#ifdef TM

#if __cplusplus
 extern "C" {
#endif

void     sp_call(void);
BOOLEAN  sp(void);

#if __cplusplus
}
#endif


BOOLEAN _cdecl sp(void)
{
   if(IS_LOCKED(ct))
      {
      UNLOCK(ct);
      sp_call();
      }
   else
      {
      sp_call();
      UNLOCK(ct);
      }
   return(TRUE);
}

#endif

